@@ -28,6 +28,8 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <ctype.h>
+#include <grp.h>
+#include <pwd.h>
 #include <sys/wait.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -47,6 +49,44 @@
 
 #include "cups.h"
 
+#if (!(CUPS_VERSION_MAJOR > 1) || (CUPS_VERSION_MINOR > 5))
+#define ippGetCount(attr)     attr->num_values
+#define ippGetGroupTag(attr)  attr->group_tag
+#define ippGetValueTag(attr)  attr->value_tag
+#define ippGetName(attr)      attr->name
+#define ippGetStatusCode(ipp) ipp->request.status.status_code
+#define ippGetString(attr, element, language) attr->values[element].string.text
+
+static ipp_attribute_t *
+ippFirstAttribute(ipp_t *ipp)
+{
+  if (!ipp)
+    return NULL;
+
+  return (ipp->current = ipp->attrs);
+}
+
+static ipp_attribute_t *
+ippNextAttribute(ipp_t *ipp)
+{
+  if (!ipp || !ipp->current)
+    return NULL;
+
+  return (ipp->current = ipp->current->next);
+}
+
+static int
+ippSetString(ipp_t            *ipp,
+             ipp_attribute_t **attr,
+             int               element,
+             const char       *strvalue)
+{
+  (*attr)->values[element].string.text = (char *) strvalue;
+
+  return 1;
+}
+#endif
+
 /* This is 0.1 second */
 #define RECONNECT_DELAY        100000
 /* We try to reconnect during 3 seconds. It's still a fairly long time even for
@@ -287,23 +327,25 @@ _cph_cups_is_printer_name_valid_internal (const char *name)
         int i;
         int len;
 
+        /* Quoting http://www.cups.org/documentation.php/doc-1.1/sam.html#4_1:
+         *
+         *    The printer name must start with any printable character except
+         *    " ", "/", and "@". It can contain up to 127 letters, numbers, and
+         *    the underscore (_).
+         *
+         * The first part is a bit weird, as the second part is more
+         * restrictive. So we only consider the second part. */
+
         /* no empty string */
         if (!name || name[0] == '\0')
                 return FALSE;
 
         len = strlen (name);
-        /* no string that is too long; see comment at the beginning of the
-         * validation code block */
-        if (len > CPH_STR_MAXLEN)
+        if (len > 127)
                 return FALSE;
 
-        /* only printable characters, no space, no /, no # */
         for (i = 0; i < len; i++) {
-                if (!g_ascii_isprint (name[i]))
-                        return FALSE;
-                if (g_ascii_isspace (name[i]))
-                        return FALSE;
-                if (name[i] == '/' || name[i] == '#')
+                if (!g_ascii_isalnum (name[i]) && name[i] != '_')
                         return FALSE;
         }
 
@@ -470,14 +512,96 @@ _CPH_CUPS_IS_VALID (filename, "filename", TRUE, CPH_STR_MAXLEN)
  * Helpers
  ******************************************************/
 
+static gboolean
+_cph_cups_set_effective_id (unsigned int   sender_uid,
+                            int           *saved_ngroups,
+                            gid_t        **saved_groups)
+{
+        struct passwd *password_entry;
+        int            ngroups;
+        gid_t         *groups;
+
+        /* avoid g_assert() because we don't want to crash here */
+        if (saved_ngroups == NULL || saved_groups == NULL) {
+                g_critical ("Internal error: cannot save supplementary groups.");
+                return FALSE;
+        }
+
+        *saved_ngroups = -1;
+        *saved_groups = NULL;
+
+        ngroups = getgroups (0, NULL);
+        if (ngroups < 0)
+                return FALSE;
+
+        groups = g_new (gid_t, ngroups);
+        if (groups == NULL && ngroups > 0)
+                return FALSE;
+
+        if (getgroups (ngroups, groups) < 0) {
+                g_free (groups);
+
+                return FALSE;
+        }
+
+        password_entry = getpwuid ((uid_t) sender_uid);
+
+        if (password_entry == NULL ||
+            setegid (password_entry->pw_gid) != 0) {
+                g_free (groups);
+
+                return FALSE;
+        }
+
+        if (initgroups (password_entry->pw_name,
+                        password_entry->pw_gid) != 0) {
+                if (getgid () != getegid ())
+                        setegid (getgid ());
+
+                g_free (groups);
+
+                return FALSE;
+        }
+
+
+        if (seteuid (sender_uid) != 0) {
+                if (getgid () != getegid ())
+                        setegid (getgid ());
+
+                setgroups (ngroups, groups);
+                g_free (groups);
+
+                return FALSE;
+        }
+
+        *saved_ngroups = ngroups;
+        *saved_groups = groups;
+
+        return TRUE;
+}
+
+static void
+_cph_cups_reset_effective_id (int    saved_ngroups,
+                              gid_t *saved_groups)
+{
+        seteuid (getuid ());
+        setegid (getgid ());
+        if (saved_ngroups >= 0)
+                setgroups (saved_ngroups, saved_groups);
+}
+
 static void
 _cph_cups_add_printer_uri (ipp_t      *request,
                            const char *name)
 {
-        char uri[HTTP_MAX_URI + 1];
+        char *escaped_name;
+        char  uri[HTTP_MAX_URI + 1];
 
+        escaped_name = g_uri_escape_string (name, NULL, FALSE);
         g_snprintf (uri, sizeof (uri),
-                    "ipp://localhost/printers/%s", name);
+                    "ipp://localhost/printers/%s", escaped_name);
+        g_free (escaped_name);
+
         ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
                       "printer-uri", NULL, uri);
 }
@@ -486,10 +610,14 @@ static void
 _cph_cups_add_class_uri (ipp_t      *request,
                          const char *name)
 {
-        char uri[HTTP_MAX_URI + 1];
+        char *escaped_name;
+        char  uri[HTTP_MAX_URI + 1];
 
+        escaped_name = g_uri_escape_string (name, NULL, FALSE);
         g_snprintf (uri, sizeof (uri),
-                    "ipp://localhost/classes/%s", name);
+                    "ipp://localhost/classes/%s", escaped_name);
+        g_free (escaped_name);
+
         ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
                       "printer-uri", NULL, uri);
 }
@@ -550,7 +678,7 @@ _cph_cups_set_error_from_reply (CphCups *cups,
                                 ipp_t   *reply)
 {
         if (reply)
-                cups->priv->last_status = reply->request.status.status_code;
+                cups->priv->last_status = ippGetStatusCode (reply);
         else
                 cups->priv->last_status = cupsLastError ();
 }
@@ -563,7 +691,7 @@ _cph_cups_is_reply_ok (CphCups  *cups,
         /* reset the internal status: we'll use the cups status */
         _cph_cups_set_internal_status (cups, NULL);
 
-        if (reply && reply->request.status.status_code <= IPP_OK_CONFLICT) {
+        if (reply && ippGetStatusCode (reply) <= IPP_OK_CONFLICT) {
                 cups->priv->last_status = IPP_OK;
                 return TRUE;
         } else {
@@ -757,28 +885,28 @@ _cph_cups_send_new_job_attributes_request (CphCups     *cups,
 }
 
 static const char *
-_cph_cups_get_attribute_string (ipp_attribute_t *attrs,
+_cph_cups_get_attribute_string (ipp_t           *reply,
                                 ipp_tag_t        group,
                                 const char      *name,
                                 ipp_tag_t        type)
 {
         ipp_attribute_t *attr;
 
-        for (attr = attrs; attr; attr = attr->next) {
-                while (attr && attr->group_tag != group)
-                        attr = attr->next;
+        for (attr = ippFirstAttribute (reply); attr; attr = ippNextAttribute (reply)) {
+                while (attr && ippGetGroupTag (attr) != group)
+                        attr = ippNextAttribute (reply);
 
                 if (attr == NULL)
                         break;
 
-                while (attr && attr->group_tag == group) {
-                        if (attr->name &&
-                            strcmp (attr->name, name) == 0 &&
-                            attr->value_tag == type) {
-                                return attr->values[0].string.text;
+                while (attr && ippGetGroupTag (attr) == group) {
+                        if (ippGetName (attr) &&
+                            strcmp (ippGetName (attr), name) == 0 &&
+                            ippGetValueTag (attr) == type) {
+                                return ippGetString (attr, 0, NULL);
                         }
 
-                        attr = attr->next;
+                        attr = ippNextAttribute (reply);
                 }
 
                 if (attr == NULL)
@@ -822,8 +950,8 @@ _cph_cups_class_has_printer (CphCups     *cups,
         if (!printer_names)
                 goto out;
 
-        for (i = 0; i < printer_names->num_values; i++) {
-                if (!g_ascii_strcasecmp (printer_names->values[i].string.text,
+        for (i = 0; i < ippGetCount (printer_names); i++) {
+                if (!g_ascii_strcasecmp (ippGetString (printer_names, i, NULL),
                                          printer_name)) {
                         retval = i;
                         break;
@@ -867,7 +995,7 @@ _cph_cups_printer_class_set_users (CphCups           *cups,
         attr = ippAddStrings (request, IPP_TAG_PRINTER, IPP_TAG_NAME,
                               request_name, len ? len : 1, NULL, NULL);
         if (len == 0)
-                attr->values[0].string.text = g_strdup (default_value);
+                ippSetString (request, &attr, 0, g_strdup (default_value));
         else {
                 int i, j;
                 for (i = 0, j = 0; i < real_len && j < len; i++) {
@@ -875,7 +1003,7 @@ _cph_cups_printer_class_set_users (CphCups           *cups,
                         if (users[i][0] == '\0')
                                 continue;
 
-                        attr->values[j].string.text = g_strdup (users[i]);
+                        ippSetString (request, &attr, j, g_strdup (users[i]));
                         j++;
                 }
         }
@@ -893,7 +1021,7 @@ _cph_cups_printer_class_set_users (CphCups           *cups,
         attr = ippAddStrings (request, IPP_TAG_PRINTER, IPP_TAG_NAME,
                               request_name, len ? len : 1, NULL, NULL);
         if (len == 0)
-                attr->values[0].string.text = g_strdup (default_value);
+                ippSetString (request, &attr, 0, g_strdup (default_value));
         else {
                 int i, j;
                 for (i = 0, j = 0; i < real_len && j < len; i++) {
@@ -901,7 +1029,7 @@ _cph_cups_printer_class_set_users (CphCups           *cups,
                         if (users[i][0] == '\0')
                                 continue;
 
-                        attr->values[j].string.text = g_strdup (users[i]);
+                        ippSetString (request, &attr, j, g_strdup (users[i]));
                         j++;
                 }
         }
@@ -992,7 +1120,7 @@ cph_cups_printer_get_uri (CphCups    *cups,
         if (!_cph_cups_is_reply_ok (cups, reply, TRUE))
                 return NULL;
 
-        const_uri = _cph_cups_get_attribute_string (reply->attrs, IPP_TAG_PRINTER,
+        const_uri = _cph_cups_get_attribute_string (reply, IPP_TAG_PRINTER,
                                                     attrs[0], IPP_TAG_URI);
 
         uri = NULL;
@@ -1033,14 +1161,17 @@ cph_cups_is_printer_local (CphCups    *cups,
 }
 
 gboolean
-cph_cups_file_get (CphCups    *cups,
-                   const char *resource,
-                   const char *filename)
+cph_cups_file_get (CphCups      *cups,
+                   const char   *resource,
+                   const char   *filename,
+                   unsigned int  sender_uid)
 {
+        int           saved_ngroups = -1;
+        gid_t        *saved_groups = NULL;
         http_status_t status;
+        int           fd;
         struct stat   file_stat;
-        uid_t         uid;
-        gid_t         gid;
+        char         *error;
 
         g_return_val_if_fail (CPH_IS_CUPS (cups), FALSE);
 
@@ -1049,44 +1180,87 @@ cph_cups_file_get (CphCups    *cups,
         if (!_cph_cups_is_filename_valid (cups, filename))
                 return FALSE;
 
-        stat (filename, &file_stat);
-        uid = file_stat.st_uid;
-        gid = file_stat.st_gid;
+        if (!_cph_cups_set_effective_id (sender_uid,
+                                         &saved_ngroups, &saved_groups)) {
+                error = g_strdup_printf ("Cannot check if \"%s\" is "
+                                         "writable: %s",
+                                         filename, strerror (errno));
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                return FALSE;
+        }
+
+        fd = open (filename, O_WRONLY | O_NOFOLLOW | O_TRUNC);
+
+        _cph_cups_reset_effective_id (saved_ngroups, saved_groups);
+        g_free (saved_groups);
+
+        if (fd < 0) {
+                error = g_strdup_printf ("Cannot open \"%s\": %s",
+                                         filename, strerror (errno));
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                return FALSE;
+        }
+
+
+        if (fstat (fd, &file_stat) != 0) {
+                error = g_strdup_printf ("Cannot write to \"%s\": %s",
+                                         filename, strerror (errno));
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                close (fd);
+
+                return FALSE;
+        }
+
+        if (!S_ISREG (file_stat.st_mode)) {
+                /* hrm, this looks suspicious... we won't help */
+                error = g_strdup_printf ("File \"%s\" is not a regular file.",
+                                         filename);
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                close (fd);
+
+                return FALSE;
+        }
 
         /* reset the internal status: we'll use the http status */
         _cph_cups_set_internal_status (cups, NULL);
 
-        status = cupsGetFile (cups->priv->connection, resource, filename);
+        status = cupsGetFd (cups->priv->connection, resource, fd);
 
         /* FIXME: There's a bug where the cups connection can fail with EPIPE.
-         * We're work-arounding it here until it's fixed in cups. */
+         * We're working around it here until it's fixed in cups. */
         if (status != HTTP_OK) {
-                if (cph_cups_reconnect (cups)) {
-                        int fd;
-
-                        /* if cupsGetFile fail, then filename is unlinked */
-                        fd = open (filename, O_CREAT, S_IRUSR | S_IWUSR);
-                        close (fd);
-                        chown (filename, uid, gid);
-
-                        _cph_cups_set_internal_status (cups, NULL);
-
-                        status = cupsGetFile (cups->priv->connection,
-                                              resource, filename);
-                }
+                if (cph_cups_reconnect (cups))
+                        status = cupsGetFd (cups->priv->connection,
+                                            resource, fd);
         }
 
+        close (fd);
+
         _cph_cups_set_internal_status_from_http (cups, status);
 
         return (status == HTTP_OK);
 }
 
 gboolean
-cph_cups_file_put (CphCups    *cups,
-                   const char *resource,
-                   const char *filename)
+cph_cups_file_put (CphCups      *cups,
+                   const char   *resource,
+                   const char   *filename,
+                   unsigned int  sender_uid)
 {
+        int           saved_ngroups = -1;
+        gid_t        *saved_groups = NULL;
         http_status_t status;
+        int           fd;
+        struct stat   file_stat;
+        char         *error;
 
         g_return_val_if_fail (CPH_IS_CUPS (cups), FALSE);
 
@@ -1095,10 +1269,60 @@ cph_cups_file_put (CphCups    *cups,
         if (!_cph_cups_is_filename_valid (cups, filename))
                 return FALSE;
 
+        if (!_cph_cups_set_effective_id (sender_uid,
+                                         &saved_ngroups, &saved_groups)) {
+                error = g_strdup_printf ("Cannot check if \"%s\" is "
+                                         "readable: %s",
+                                         filename, strerror (errno));
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                return FALSE;
+        }
+
+        fd = open (filename, O_RDONLY);
+
+        _cph_cups_reset_effective_id (saved_ngroups, saved_groups);
+        g_free (saved_groups);
+
+        if (fd < 0) {
+                error = g_strdup_printf ("Cannot open \"%s\": %s",
+                                         filename, strerror (errno));
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                return FALSE;
+        }
+
+        if (fstat (fd, &file_stat) != 0) {
+                error = g_strdup_printf ("Cannot read \"%s\": %s",
+                                         filename, strerror (errno));
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                close (fd);
+
+                return FALSE;
+        }
+
+        if (!S_ISREG (file_stat.st_mode)) {
+                /* hrm, this looks suspicious... we won't help */
+                error = g_strdup_printf ("File \"%s\" is not a regular file.",
+                                         filename);
+                _cph_cups_set_internal_status (cups, error);
+                g_free (error);
+
+                close (fd);
+
+                return FALSE;
+        }
+
         /* reset the internal status: we'll use the http status */
         _cph_cups_set_internal_status (cups, NULL);
 
-        status = cupsPutFile (cups->priv->connection, resource, filename);
+        status = cupsPutFd (cups->priv->connection, resource, fd);
+
+        close (fd);
 
         _cph_cups_set_internal_status_from_http (cups, status);
 
@@ -1351,7 +1575,7 @@ _cph_cups_devices_get_old (CphCups           *cups,
                 attr = ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
                                       "include-schemes", len_include, NULL, NULL);
                 for (i = 0; i < len_include; i++)
-                        attr->values[i].string.text = g_strdup (include_schemes[i]);
+                        ippSetString (request, &attr, i, g_strdup (include_schemes[i]));
         }
 
         if (exclude_schemes && len_exclude > 0) {
@@ -1360,7 +1584,7 @@ _cph_cups_devices_get_old (CphCups           *cups,
                 attr = ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
                                       "exclude-schemes", len_exclude, NULL, NULL);
                 for (i = 0; i < len_exclude; i++)
-                        attr->values[i].string.text = g_strdup (exclude_schemes[i]);
+                        ippSetString (request, &attr, i, g_strdup (exclude_schemes[i]));
         }
 
         resource_char = _cph_cups_get_resource (CPH_RESOURCE_ROOT);
@@ -1370,9 +1594,9 @@ _cph_cups_devices_get_old (CphCups           *cups,
         if (!_cph_cups_is_reply_ok (cups, reply, TRUE))
                 return FALSE;
 
-        for (attr = reply->attrs; attr; attr = attr->next) {
-                while (attr && attr->group_tag != IPP_TAG_PRINTER)
-                        attr = attr->next;
+        for (attr = ippFirstAttribute (reply); attr; attr = ippNextAttribute (reply)) {
+                while (attr && ippGetGroupTag (attr) != IPP_TAG_PRINTER)
+                        attr = ippNextAttribute (reply);
 
                 if (attr == NULL)
                         break;
@@ -1384,29 +1608,29 @@ _cph_cups_devices_get_old (CphCups           *cups,
                 device_make_and_model = NULL;
                 device_uri            = NULL;
 
-                while (attr && attr->group_tag == IPP_TAG_PRINTER) {
-                        if (attr->name == NULL)
+                while (attr && ippGetGroupTag (attr) == IPP_TAG_PRINTER) {
+                        if (ippGetName (attr) == NULL)
                                 /* nothing, just skip */;
-                        else if (strcmp (attr->name, "device-class") == 0 &&
-                                 attr->value_tag == IPP_TAG_KEYWORD)
-                                device_class = g_strdup (attr->values[0].string.text);
-                        else if (strcmp (attr->name, "device-id") == 0 &&
-                                 attr->value_tag == IPP_TAG_TEXT)
-                                device_id = g_strdup (attr->values[0].string.text);
-                        else if (strcmp (attr->name, "device-info") == 0 &&
-                                 attr->value_tag == IPP_TAG_TEXT)
-                                device_info = g_strdup (attr->values[0].string.text);
-                        else if (strcmp (attr->name, "device-location") == 0 &&
-                                 attr->value_tag == IPP_TAG_TEXT)
-                                device_location = g_strdup (attr->values[0].string.text);
-                        else if (strcmp (attr->name, "device-make-and-model") == 0 &&
-                                 attr->value_tag == IPP_TAG_TEXT)
-                                device_make_and_model = g_strdup (attr->values[0].string.text);
-                        else if (strcmp (attr->name, "device-uri") == 0 &&
-                                 attr->value_tag == IPP_TAG_URI)
-                                device_uri = g_strdup (attr->values[0].string.text);
+                        else if (strcmp (ippGetName (attr), "device-class") == 0 &&
+                                 ippGetValueTag (attr) == IPP_TAG_KEYWORD)
+                                device_class = g_strdup (ippGetString (attr, 0, NULL));
+                        else if (strcmp (ippGetName (attr), "device-id") == 0 &&
+                                 ippGetValueTag (attr) == IPP_TAG_TEXT)
+                                device_id = g_strdup (ippGetString (attr, 0, NULL));
+                        else if (strcmp (ippGetName (attr), "device-info") == 0 &&
+                                 ippGetValueTag (attr) == IPP_TAG_TEXT)
+                                device_info = g_strdup (ippGetString (attr, 0, NULL));
+                        else if (strcmp (ippGetName (attr), "device-location") == 0 &&
+                                 ippGetValueTag (attr) == IPP_TAG_TEXT)
+                                device_location = g_strdup (ippGetString (attr, 0, NULL));
+                        else if (strcmp (ippGetName (attr), "device-make-and-model") == 0 &&
+                                 ippGetValueTag (attr) == IPP_TAG_TEXT)
+                                device_make_and_model = g_strdup (ippGetString (attr, 0, NULL));
+                        else if (strcmp (ippGetName (attr), "device-uri") == 0 &&
+                                 ippGetValueTag (attr) == IPP_TAG_URI)
+                                device_uri = g_strdup (ippGetString (attr, 0, NULL));
 
-                        attr = attr->next;
+                        attr = ippNextAttribute (reply);
                 }
 
                 if (device_uri)
@@ -1702,6 +1926,7 @@ cph_cups_class_add_printer (CphCups    *cups,
         ipp_t           *request;
         int              new_len;
         ipp_attribute_t *printer_uris;
+        char            *escaped_printer_name;
         char             printer_uri[HTTP_MAX_URI + 1];
         ipp_attribute_t *attr;
 
@@ -1736,8 +1961,10 @@ cph_cups_class_add_printer (CphCups    *cups,
         _cph_cups_add_class_uri (request, class_name);
         _cph_cups_add_requesting_user_name (request, NULL);
 
+        escaped_printer_name = g_uri_escape_string (printer_name, NULL, FALSE);
         g_snprintf (printer_uri, sizeof (printer_uri),
-                    "ipp://localhost/printers/%s", printer_name);
+                    "ipp://localhost/printers/%s", escaped_printer_name);
+        g_free (escaped_printer_name);
 
         /* new length: 1 + what we had before */
         new_len = 1;
@@ -1745,7 +1972,7 @@ cph_cups_class_add_printer (CphCups    *cups,
                 printer_uris = ippFindAttribute (reply,
                                                  "member-uris", IPP_TAG_URI);
                 if (printer_uris)
-                        new_len += printer_uris->num_values;
+                        new_len += ippGetCount (printer_uris);
         } else
                 printer_uris = NULL;
 
@@ -1755,14 +1982,15 @@ cph_cups_class_add_printer (CphCups    *cups,
         if (printer_uris) {
                 int i;
 
-                for (i = 0; i < printer_uris->num_values; i++)
-                        attr->values[i].string.text = g_strdup (printer_uris->values[i].string.text);
+                for (i = 0; i < ippGetCount (printer_uris); i++)
+                        ippSetString (request, &attr, i,
+                                      g_strdup (ippGetString (printer_uris, i, NULL)));
         }
 
         if (reply)
                 ippDelete (reply);
 
-        attr->values[new_len - 1].string.text = g_strdup (printer_uri);
+        ippSetString (request, &attr, new_len - 1, g_strdup (printer_uri));
 
         return _cph_cups_send_request (cups, request, CPH_RESOURCE_ADMIN);
 }
@@ -1814,7 +2042,7 @@ cph_cups_class_delete_printer (CphCups    *cups,
         printer_uris = ippFindAttribute (reply,
                                          "member-uris", IPP_TAG_URI);
         if (printer_uris)
-                new_len += printer_uris->num_values;
+                new_len += ippGetCount (printer_uris);
 
         /* empty class: we delete it */
         if (new_len <= 0) {
@@ -1834,9 +2062,11 @@ cph_cups_class_delete_printer (CphCups    *cups,
 
         /* copy all printers from the class, except the one we remove */
         for (i = 0; i < printer_index; i++)
-                attr->values[i].string.text = g_strdup (printer_uris->values[i].string.text);
-        for (i = printer_index + 1; i < printer_uris->num_values; i++)
-                attr->values[i].string.text = g_strdup (printer_uris->values[i].string.text);
+                ippSetString (request, &attr, i,
+                              g_strdup (ippGetString (printer_uris, i, NULL)));
+        for (i = printer_index + 1; i < ippGetCount (printer_uris); i++)
+                ippSetString (request, &attr, i,
+                              g_strdup (ippGetString (printer_uris, i, NULL)));
 
         ippDelete (reply);
 
@@ -2133,7 +2363,7 @@ cph_cups_printer_class_set_option_default (CphCups           *cups,
                                       option_name, len, NULL, NULL);
 
                 for (i = 0; i < len; i++)
-                        attr->values[i].string.text = g_strdup (values[i]);
+                        ippSetString (request, &attr, i, g_strdup (values[i]));
         }
 
         retval = _cph_cups_send_request (cups, request, CPH_RESOURCE_ADMIN);
@@ -2358,7 +2588,7 @@ cph_cups_printer_class_set_option (CphCups           *cups,
                                       option, len, NULL, NULL);
 
                 for (i = 0; i < len; i++)
-                        attr->values[i].string.text = g_strdup (values[i]);
+                        ippSetString (request, &attr, i, g_strdup (values[i]));
         }
 
         if (newppdfile) {
@@ -2477,7 +2707,7 @@ cph_cups_job_get_status (CphCups    *cups,
         if (!_cph_cups_is_reply_ok (cups, reply, TRUE))
                 return CPH_JOB_STATUS_INVALID;
 
-        orig_user = _cph_cups_get_attribute_string (reply->attrs, IPP_TAG_JOB,
+        orig_user = _cph_cups_get_attribute_string (reply, IPP_TAG_JOB,
                                                     attrs[0], IPP_TAG_NAME);
 
         status = CPH_JOB_STATUS_INVALID;
